;________________________________________________________________________
;
;                            T I M I N G . A S M
;________________________________________________________________________
;
; PIC16F84/16C73A project by Byon Garrabrant  N6BG  byon@mail.com  Apr 98
;________________________________________________________________________

#define	PTT_IN		PORTB, 4
#define	VOX_IN		PORTB, 5
#define	VALID_OUT	PORTA, 3
#define	VALID_TIME	D'15'
;________________________________________________________________________

	cblock
		lastTMR0
		sub_second_timer
		ten_second_timer
		quiet_timer
 		transmission_timer
		ptt_in_history
	endc
;________________________________________________________________________

Timing_Init
	movlw	D'1'
	movwf	ten_second_timer
	movwf	sub_second_timer
 	clrf	NMEA_valid_timer	; init valid timer to 255
 	decf	NMEA_valid_timer, F
	clrf	quiet_timer
	clrf	transmission_timer
	clrf	ptt_in_history
	return

Timing_Check_Update
	clrc				; update ptt history
	btfsc	PTT_IN
	 setc
	rlf	ptt_in_history, F
	btfsc	VOX_IN			; clear quiet_timer
	 clrf	quiet_timer		; if voice activity

	movfw	lastTMR0		; get difference from
	subwf	TMR0, W			; last time and
	addwf	lastTMR0, F		; add to last time
	skpc				; if no wrap
	 return				; leave
	decfsz	sub_second_timer, F	; count down the sub seconds
	 return				; if not zero, leave

	movlw	D'38'			; reset the sub seconds
;	movlw	D'4'			; TEN X SPEEDUP
	movwf	sub_second_timer	
	incf	quiet_timer, F		; bump quiet_timer
	skpnz				; but clamp
	 decf	quiet_timer, F		; at 255
	btfsc	VOX_IN			; clear quiet_timer
	 clrf	quiet_timer		; if voice activity
	decfsz	ten_second_timer, F	; count down ten seconds
	 return				; if not done, leave

	movlw	D'10'			; reset the ten second timer
	movwf	ten_second_timer
	incf	transmission_timer, F	; inc transmission_timer
	skpnz				; but clamp
	 decf	transmission_timer, F	; to 255

	incf	NMEA_valid_timer, F	; inc valid_timer
	skpnz				; but clamp
	 decf	NMEA_valid_timer, F	; to 255

	return
	

Timing_Check
	call	Timing_Check_Update

	movfw	NMEA_valid_timer	; update the valid light
	sublw	VALID_TIME
	skpc
	 bcf	VALID_OUT
	skpnc
	 bsf	VALID_OUT


	movlw	OPT_AUTO_TIME		; check auto time
	call	Get_Option
	subwf	transmission_timer, W
	bnc	Check_Manual

	movlw	OPT_QUIET_TIME		; check quiet time
	call	Get_Option
	subwf	quiet_timer, W
	bc	Do_Transmit

Check_Manual
	movlw	OPT_MANUAL_TIME		; check manual time
	call	Get_Option
	subwf	transmission_timer, W
	bnc	Check_Done

	movfw	ptt_in_history
	andlw	H'3'
	xorlw	H'2'
	bnz	Check_Done

Do_Transmit
	clrf	transmission_timer
	call	NMEA_Reset
	retlw	1

Check_Done	
	retlw	0

;________________________________________________________________________
